#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Test database integrity. Try to catch weird errors and fix them.
'''

import sys, argparse, locale, signal, logging, sqlite3

__version__ = '1.0.0'
DB_FILE = '/var/lib/tuptime/tuptime.db'
fixcnt = 0
errcnt = 0

# List of tests to auto-execute
TESTS = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

# Terminate when SIGPIPE signal is received
signal.signal(signal.SIGPIPE, signal.SIG_DFL)

# Set locale to the users default settings (LANG env. var)
locale.setlocale(locale.LC_ALL, '')


def get_arguments():
    """Get arguments from command line"""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-f', '--filedb',
        dest='db_file',
        default=DB_FILE,
        action='store',
        help='database file (' + DB_FILE + ')',
        metavar='FILE'
    )
    parser.add_argument(
        '--fix',
        dest='fix',
        default=False,
        action='store_true',
        help='try to fix errors automatically'
    )
    parser.add_argument(
        '-t', '--test',
        dest='test',
        nargs='+',
        type=int,
        default=TESTS,
        help='execute only this test'
    )
    parser.add_argument(
        '-v', '--verbose',
        dest='verbose',
        default=False,
        action='store_true',
        help='verbose output'
    )
    arg = parser.parse_args()

    if arg.verbose:
        logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
        logging.info('Version: %s', (__version__))

    logging.info('Arguments: %s', str(vars(arg)))
    return arg


def err_cnt(arg):
    global errcnt
    global fixcnt
    if arg.fix:
        fixcnt += 1
    errcnt += 1


def test0(arg, db_rows, conn):
    if len(db_rows) != db_rows[-1]['startup']:
        print(' Possible deleted rows in db. Real startups are not equal to enumerate startups')

        if arg.fix:
            conn.execute('vacuum')
            print(' FIXED: vacuum')
        err_cnt(arg)


def test1(arg, row, conn):
    if row['offbtime'] and \
       row['btime'] > row['offbtime']:
        print(row['startup'])
        print(' row btime > offbtime')
        print(' ' + str(row['btime']) + ' > ' + str(row['offbtime']))

        if arg.fix:
            conn.execute('delete from tuptime where rowid = ' + str(row['startup']))
            print(' FIXED: delete row = ' + str(row['startup']))
        err_cnt(arg)


def test2(arg, row, conn, prev_row):
    if prev_row['offbtime'] > row['btime']:
        print(row['startup'])
        print(' prev_row offbtime > btime')
        print(' ' + str(prev_row['offbtime']) + ' > ' + str(row['btime']))

        if arg.fix:
            conn.execute('delete from tuptime where rowid = ' + str(row['startup']))
            print(' FIXED: delete row = ' + str(row['startup']))
        err_cnt(arg)


def test3(arg, row, conn, prev_row):
    if prev_row['offbtime'] + prev_row['downtime'] != row['btime']:
        print(row['startup'])
        print(' prev_row offbtime + prev_row downtime != btime')
        print(' ' + str(prev_row['offbtime']) + ' + ' + str(prev_row['downtime']) + ' != ' + str(row['btime']))

        if arg.fix:
            fixed = row['btime'] - prev_row['offbtime']
            conn.execute('update tuptime set downtime = ' + str(fixed) + ' where rowid = ' + str(row['startup'] - 1))
            print(' FIXED: prev_row downtime = ' + str(fixed))
        err_cnt(arg)


def test4(arg, row, conn):
    if row['offbtime'] and \
       row['btime'] + row['uptime'] != row['offbtime']:
        print(row['startup'])
        print(' row btime + uptime != offbtime')
        print(' ' + str(row['btime']) + ' + ' + str(row['uptime']) + ' != ' + str(row['offbtime']))

        if arg.fix:
            fixed = row['offbtime'] - row['btime']
            conn.execute('update tuptime set uptime = ' + str(fixed) + ' where rowid = ' + str(row['startup']))
            print(' FIXED: uptime = ' + str(fixed))
        err_cnt(arg)


def test5(arg, row, conn):
    if row['rntime'] + row['slptime'] != row['uptime']:
        print(row['startup'])
        print(' rntime + slptime != uptime')
        print(' ' + str(row['rntime']) + ' + ' + str(row['slptime']) + ' != ' + str(row['uptime']))

        if arg.fix:
            fixed = row['rntime'] + row['slptime'] - row['uptime']
            if row['rntime'] > row['slptime'] and row['rntime'] - fixed > 0:
                fixed2 = row['rntime'] - fixed
                conn.execute('update tuptime set rntime = ' + str(fixed2) + ' where rowid = ' + str(row['startup']))
                print(' FIXED: rntime = ' + str(fixed2))
            elif row['slptime'] > row['rntime'] and row['slptime'] - fixed > 0:
                fixed2 = row['slptime'] - fixed
                conn.execute('update tuptime set slptime = ' + str(fixed2) + ' where rowid = ' + str(row['startup']))
                print(' FIXED: slptime = ' + str(fixed2))
            else:
                conn.execute('update tuptime set rntime = ' + str(row['uptime']) + ', slptime = 0 where rowid = ' + str(row['startup']))
                print(' FIXED: rntime = ' + str(row['uptime']))
                print(' FIXED: slptime = 0')
        err_cnt(arg)


def test6(arg, row, conn):
    if row['uptime'] < 0:
        print(row['startup'])
        print(' uptime < 0')
        print(' ' + str(row['uptime']) + ' < 0')

        if arg.fix:
            conn.execute('delete from tuptime where rowid = ' + str(row['startup']))
            print(' FIXED: delete row = ' + str(row['startup']))
        err_cnt(arg)


def test7(arg, row, conn):
    if row['rntime'] < 0:
        print(row['startup'])
        print(' rntime < 0')
        print(' ' + str(row['rntime']) + ' < 0')

        if arg.fix:
            conn.execute('delete from tuptime where rowid = ' + str(row['startup']))
            print(' FIXED: delete row = ' + str(row['startup']))
        err_cnt(arg)


def test8(arg, row, conn):
    if row['slptime'] < 0:
        print(row['startup'])
        print(' slptime < 0')
        print(' ' + str(row['slptime']) + ' < 0')

        if arg.fix:
            conn.execute('delete from tuptime where rowid = ' + str(row['startup']))
            print(' FIXED: delete row = ' + str(row['startup']))
        err_cnt(arg)


def test9(arg, row, conn):
    if row['downtime'] and \
       row['downtime'] < 0:
        print(row['startup'])
        print(' downtime < 0')
        print(' ' + str(row['downtime']) + ' < 0')

        if arg.fix:
            conn.execute('delete from tuptime where rowid = ' + str(row['startup']))
            print(' FIXED: delete row = ' + str(row['startup']))
        err_cnt(arg)



def main():

    arg = get_arguments()

    db_conn = sqlite3.connect(arg.db_file)
    db_conn.row_factory = sqlite3.Row
    conn = db_conn.cursor()

    # Check if DB have the old format
    columns = [i[1] for i in conn.execute('PRAGMA table_info(tuptime)')]
    if 'rntime' and 'slptime' not in columns:
        logging.error('DB format outdated')
        sys.exit(-1)

    print('Processing ' + str(arg.db_file) + ' --->')

    for i in arg.test:
        print('\n### Test ' + str(i) + ' ###')

        conn.execute('select rowid as startup, * from tuptime')
        db_rows = conn.fetchall()
        for row in db_rows:

            if arg.verbose:
                print('\n' + str(row['startup']) + '\n ' + str(dict(row)))

            if i == 1:
                test1(arg, row, conn)

            if i == 2:
                if row != db_rows[0]:  # Only after first row
                    test2(arg, row, conn, prev_row)

            if i == 3:
                if row != db_rows[0]:  # Only after first row
                    test3(arg, row, conn, prev_row)

            if i == 4:
                test4(arg, row, conn)

            if i == 5:
                test5(arg, row, conn)

            if i == 6:
                test6(arg, row, conn)

            if i == 7:
                test7(arg, row, conn)

            if i == 8:
                test8(arg, row, conn)

            if i == 9:
                test9(arg, row, conn)

            if i == 9:
                test9(arg, row, conn)

            prev_row = row

        if i == 10:
            test0(arg, db_rows, conn)

        db_conn.commit()

    db_conn.close()

    print('\n' + '-' * 25)
    print('Errors: ' + str(errcnt))
    print('Fixed: ' + str(fixcnt))
    print('')


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8

import sqlite3
import copy
import argparse
import sys
import logging
import matplotlib.pyplot as plt
from datetime import datetime, timedelta

SECONDS_DAY = 86400


class UptimeRangeInDay:
    # at least 3 splits in each day, represent
    # startup, shutdown and bad shutdown states.
    max_split_count = 3

    def __init__(self, btime, uptime):
        """
        :splits: list, each element is time consumed one of 3 different computer
            state (shutdown, startup, bad shutdown) in same day, the first state
            is always shutdown, then startup, bad shutdown, shutdown...
            But the time consumed on one state can be 0(boot pc at midnight, the
            first element will be 0, as the first state is shutdown)
        :date: date of this startup, if the `btime` and `offtime` of db record
            crossing midnight, it will be split to 2 parts.
        :start_point: start time of this state, changing for different states.
        :end_point: end time of this state.
        :bad_split_idxes: index of bad shutdown state in splits.
        """
        bdate = datetime.fromtimestamp(btime)

        self._splits = []
        self._date = datetime(bdate.year, bdate.month, bdate.day)
        self._start_point = (btime - self._date.timestamp()) / SECONDS_DAY
        self._end_point = self._start_point + uptime / SECONDS_DAY
        self.add_splits((self._start_point,
                        self._end_point - self._start_point, 0))

    @property
    def splits(self):
        return self._splits

    @property
    def date(self):
        return self._date

    @property
    def start_point(self):
        return self._start_point

    @property
    def end_point(self):
        return self._end_point

    @end_point.setter
    def end_point(self, end_point):
        self._end_point = end_point

    @classmethod
    def update_max_split_count(cls, count):
        if count > cls.max_split_count:
            cls.max_split_count = count

    def add_splits(self, split_consumed_times):
        """add time cousumed into splits."""
        self._splits.extend(split_consumed_times)
        self.update_max_split_count(len(self))

    def exchange_splits(self, idx, idy):
        self._splits[idx], self._splits[idy] = \
                self._splits[idy], self._splits[idx]

    def __len__(self):
        return len(self._splits)


def get_uptime_data(arg):
    """Get all rows from database."""

    db_conn = sqlite3.connect(arg.db_file)
    db_conn.row_factory = sqlite3.Row
    conn = db_conn.cursor()
    conn.execute('select rowid as startup, * from tuptime')
    db_rows = conn.fetchall()
    db_conn.close()
    db_rows = [dict(row) for row in db_rows]

    tuptime_install_date = get_last_midnight_date(datetime.fromtimestamp(db_rows[0]['btime']))
    if arg.begin_date < tuptime_install_date:
        logging.warning(f"First tuptime entry was recorded at {tuptime_install_date:%Y-%m-%d}.")

    # last db entry is right now running state
    db_row = db_rows[-1]
    if db_row['offbtime'] == -1:
        db_row['offbtime'] = db_row['btime'] + db_row['uptime']
        db_row['endst'] = 1

    selected_db_rows = []
    # including tuptime record of day before begin_date
    # as last tuptime record may cross midnight
    begin_second = int(arg.begin_date.strftime("%s"))
    end_second = int(arg.end_date.strftime("%s"))

    for db_row in db_rows:
        if db_row['endst'] == 0:  # bad shutdown
            db_row['offbtime'] = int(db_row['offbtime'] + db_row['downtime'])
        if db_row['offbtime'] > begin_second  and  db_row['btime'] < end_second:
            selected_db_rows.append(db_row)

    if len(selected_db_rows) == 0:
        logging.warning("No tuptime entries in this date range.")
        sys.exit(1)

    return selected_db_rows


def get_uptime_range_each_day(selected_db_rows, arg):
    """Get all states for all date between bdate and edate.
    Focus on urid's splits which is a sequence of time consumed on each
    state. The previous state of a startup state must be a shotdown state
    in splits, the next state of a strtup state must be a bad shutdown
    state.
    When way read a new record, we always append 3 type consumed time to
    splits(ordered in shutdown, startup, bad shutdown).
    """

    uptime_ranges = []
    bdate_prev_record = datetime.fromtimestamp(selected_db_rows[0]['btime'])

    def create_or_update_uptime_range():
        """Create uptime range object and insert state
        start/end point to object's splits."""

        nonlocal bdate_prev_record

        if urid.date == bdate_prev_record:
            urid_prev_record = uptime_ranges[-1]
            downtime = urid.start_point - urid_prev_record.end_point
            uptime = urid.end_point - urid.start_point
            badtime = 0
            if entry_state == 0:  # bad shutdown record
                uptime, badtime = badtime, uptime
            urid_prev_record.add_splits((downtime, uptime, badtime))  # badtime
            urid_prev_record.end_point = urid.end_point
        else:
            if entry_state == 0:
                # exchange time consumed between first uptime and first badtime
                urid.exchange_splits(1, 2)
            uptime_ranges.append(urid)
            bdate_prev_record = urid.date

    for db_row in selected_db_rows:
        entry_state = db_row['endst']
        btime = db_row['btime']
        bdate = datetime.fromtimestamp(btime)
        midnight_date = datetime(bdate.year, bdate.month, bdate.day) + timedelta(days=1)
        offbtime = db_row['offbtime']
        while True:
            # split record if corssing midnight
            if offbtime > midnight_date.timestamp():
                urid = UptimeRangeInDay(btime, midnight_date.timestamp() - btime)
                btime = midnight_date.timestamp()
                if urid.date == arg.begin_date - timedelta(days=1):
                    continue
                create_or_update_uptime_range()
                # break, otherwise will cross end_date
                if midnight_date == arg.end_date:
                    break
                midnight_date += timedelta(days=1)
            else:
                urid = UptimeRangeInDay(btime, offbtime - btime)
                create_or_update_uptime_range()
                break

    # last urid index which should set last pc state
    idx_urid = len(uptime_ranges) - 1
    if arg.end_date < datetime.today():
        idx_urid += 1
    # set last pc state to shutdown, if total time of splits less than 1
    for up in uptime_ranges[:idx_urid]:
        total_time = sum(up.splits)
        if abs(1 - total_time) > 1e-3:
            up.add_splits((abs(1 - total_time),))

    if len(uptime_ranges) == 0:
        logging.warning("Computer is running, no other record in this date range in DB.")
        sys.exit(1)

    return uptime_ranges


def plot_time(uptime_ranges, arg):
    """Plot stacked bar chart."""
    max_splits_in_day = UptimeRangeInDay.max_split_count

    # different day got different BAD record, statistic all index that got bad
    # record, insert a 0 to split if that day didn't have a bad record at same
    # split index for all indexes.
    data = [[0] * len(uptime_ranges) for _ in range(max_splits_in_day)]
    for data_col, urid in enumerate(uptime_ranges):
        for data_row, split in enumerate(urid.splits):
            data[data_row][data_col] = split * 24

    bottom_data = copy.deepcopy(data)
    for i in range(1, len(bottom_data)):
        for j in range(len(bottom_data[0])):
            bottom_data[i][j] += bottom_data[i - 1][j]

    fig, ax = plt.subplots()
    fig.set_size_inches(arg.fig_width, arg.fig_height)
    ind = list(range(len(uptime_ranges)))
    width = arg.bar_width

    p1 = ax.bar(ind, data[0], width, color='b')
    p2 = ax.bar(ind, data[1], width, color='r', bottom=bottom_data[0])
    p3 = ax.bar(ind, data[2], width, color='k', bottom=bottom_data[1])
    colors = {0: 'b', 1: 'r', 2: 'k'}
    for i in range(3, max_splits_in_day):
        region_color = colors[i % 3]
        ax.bar(ind, data[i], width, color=region_color, bottom=bottom_data[i-1])

    ax.set_yticks(list(range(0, 25, 4)))
    ax.set_yticks(list(range(0, 25, 2)), minor=True)
    xticks = [f"{up.date:%Y-%m-%d}" for up in uptime_ranges]
    ax.set_xticks(ind)
    ax.set_xticklabels(xticks, rotation=arg.xtick_labels_rotation)
    ax.set_title("Tuptime bar chart")
    ax.set_ylabel('Hours')
    ax.set_xlabel('Date')
    ax.legend((p1[0], p2[0], p3[0]), ('downtime', 'uptime', 'badtime'), loc="upper right")
    ax.grid(which='minor', axis='y', linestyle=(0, (3, 10, 1, 10)),
            linewidth=arg.line_width, alpha=arg.grid_alpha)
    plt.show()


def get_last_midnight_date(date):
    return datetime.combine(date, datetime.min.time())


def get_arguments():
    DB_FILE = '/var/lib/tuptime/tuptime.db'

    class CustomHelpFormatter(argparse.HelpFormatter):
        def _format_action_invocation(self, action):
            if not action.option_strings or action.nargs == 0:
                return super()._format_action_invocation(action)
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            return ', '.join(action.option_strings) + ' ' + args_string

    parser = argparse.ArgumentParser(formatter_class=lambda prog: CustomHelpFormatter(prog))
    parser.add_argument(
        '-f', '--filedb',
        dest='db_file',
        default=DB_FILE,
        action='store',
        help='database file',
        metavar='FILE'
    )
    parser.add_argument(
        '-b', '--bdate',
        dest='begin_date',
        action='store',
        help='begin date to plot, format:Y-m-d.',
        type=str
    )
    parser.add_argument(
        '-e', '--edate',
        dest='end_date',
        action='store',
        help='end date to plot, format:Y-m-d. Default edate is today.',
        type=str
    )
    parser.add_argument(
        '-p', '--pastdays',
        dest='past_days',
        default=7,
        action='store',
        help='past days before edate to plot, will be ignored if set bdate (default is 7).',
        type=int
    )
    parser.add_argument(
        '--fwidth',
        dest='fig_width',
        default=10,
        action='store',
        help='figure width.',
        type=int
    )
    parser.add_argument(
        '--fheight',
        dest='fig_height',
        default=12,
        action='store',
        help='figure height.',
        type=int
    )
    parser.add_argument(
        '-w', '--bwidth',
        dest='bar_width',
        default=.5,
        action='store',
        help='width of the bars (default is 0.5).',
        type=float
    )
    parser.add_argument(
        '-r', '--xrotation',
        dest='xtick_labels_rotation',
        default=30,
        action='store',
        help='rotation of xticks label (default is 30 drgrees).',
        type=float
    )
    parser.add_argument(
        '--lwidth',
        dest='line_width',
        default=1,
        action='store',
        help='line width of figure grid (default is 1).',
        type=float
    )
    parser.add_argument(
        '-a', '--galpha',
        dest='grid_alpha',
        default=.0,
        action='store',
        help='alpha value of figure grid (default is 0).',
        type=float
    )

    arg = parser.parse_args()
    logging.basicConfig(format='%(levelname)s:%(message)s')

    if arg.end_date:
        arg.end_date = datetime.strptime(arg.end_date, "%Y-%m-%d") + timedelta(days=1)
        if arg.end_date > datetime.today() + timedelta(days=1):
            logging.error("end_date can't large than today.")
            sys.exit(-1)
    else:
        arg.end_date = get_last_midnight_date(datetime.today()) + timedelta(days=1)

    if arg.begin_date:
        arg.begin_date = datetime.strptime(arg.begin_date, "%Y-%m-%d")
    else:
        arg.begin_date = arg.end_date - timedelta(days=arg.past_days)

    if arg.begin_date >= arg.end_date:
        logging.error("begin_date must be earlier than end_date.")
        sys.exit(-1)

    return arg


def main():

    arg = get_arguments()
    selected_db_rows = get_uptime_data(arg)
    uptime_ranges = get_uptime_range_each_day(selected_db_rows, arg)
    plot_time(uptime_ranges, arg)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
This script join two tuptime db files into an other one. It works as follows:

    - Takes two paths of db files (old one must be first) and an output file as destination of the join.
    - Look for last register on old file and sum (btime + uptime)
    - Look on new file a btime greather than previous sum value
    - Insert registers that match

 # tuptime_join.py -d /path/to/joined.db /path/to/old.db /path/to/new.db

Maybe after upgrade your computer and install new stuff, you want to continue with the registers
that you have before. It is possible to join the new ones into the old ones without any awkward jump.
Please, see the following example:

Join new registers of two db files into other (old file must be defined first):
    tuptime_join.py -d /tmp/tt.db /backup/old/tuptime.db /var/lib/tuptime/tuptime.db

Check if all is ok:
    tuptime --noup -t -f /tmp/tt.db

Check owner (usually tuptime:tuptime) and copy modified file to right location. Re-check owner:
    ls -al /var/lib/tuptime/tuptime.db
    cp /tmp/tt.db /var/lib/tuptime/tuptime.db
    ls -al /var/lib/tuptime/tuptime.db

'''

import sys, argparse, locale, signal, logging, sqlite3
from shutil import copyfile

__version__ = '1.0.0'

# Terminate when SIGPIPE signal is received
signal.signal(signal.SIGPIPE, signal.SIG_DFL)

# Set locale to the users default settings (LANG env. var)
locale.setlocale(locale.LC_ALL, '')


def get_arguments():
    """Get arguments from command line"""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'files',
        metavar='FILE',
        nargs=2,
        type=str,
        help='files to join (OLD and NEW)'
    )
    parser.add_argument(
        '-d', '--dest',
        dest='dest',
        default=False,
        action='store',
        type=str,
        required=True,
        help='destination file to store join'
    )
    parser.add_argument(
        '-v', '--verbose',
        dest='verbose',
        default=False,
        action='store_true',
        help='verbose output'
    )
    arg = parser.parse_args()

    if arg.verbose:
        logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
        logging.info('Version: %s', (__version__))

    logging.info('Arguments: %s', str(vars(arg)))
    return arg


def main():

    arg = get_arguments()

    print('Old file (arg 1): \t' + str(arg.files[0]))
    print('New file (arg 2):\t' + str(arg.files[1]))

    # Copy old file to output file
    print('\nCopy old file to destination')
    copyfile(arg.files[0], arg.dest)
    fl0 = {'name': arg.dest}    # File0 is output file
    fl1 = {'name': arg.files[1]}  # file1 is new file
    print('Destination file: \t' + str(fl0['name']))

    # Open file0 DB
    db_conn0 = sqlite3.connect(fl0['name'])
    db_conn0.row_factory = sqlite3.Row
    conn0 = db_conn0.cursor()

    # Check if DB have the old format
    columns = [i[1] for i in conn0.execute('PRAGMA table_info(tuptime)')]
    if 'rntime' and 'slptime' not in columns:
        logging.error('DB format outdated on old file')
        sys.exit(-1)

    # Open file1 DB
    db_conn1 = sqlite3.connect(fl1['name'])
    db_conn1.row_factory = sqlite3.Row
    conn1 = db_conn1.cursor()

    # Check if DB have the old format
    columns = [i[1] for i in conn1.execute('PRAGMA table_info(tuptime)')]
    if 'rntime' and 'slptime' not in columns:
        logging.error('DB format outdated on new file')
        sys.exit(-1)

    # Check older file
    conn0.execute('select btime from tuptime where rowid = (select min(rowid) from tuptime)')
    fl0_btime = conn0.fetchone()[0]
    conn1.execute('select btime from tuptime where rowid = (select min(rowid) from tuptime)')
    fl1_btime = conn1.fetchone()[0]
    if fl0_btime > fl1_btime:
        logging.warning('"New file" looks older than "Old file"')

    # Print raw rows
    conn0.execute('select rowid as startup, * from tuptime')
    db_rows = conn0.fetchall()
    print('\nDestination rows before:\t' + str(len(db_rows)))
    if arg.verbose:
        for row in db_rows:
            print("\t", end='')
            print(dict(row))

    # Get last startup, btime and uptime from last row on file0 to calculate offbtime
    conn0.execute('select rowid, btime, uptime from tuptime where rowid = (select max(rowid) from tuptime)')
    fl0['startup'], fl0['btime'], fl0['uptime'] = conn0.fetchone()
    fl0['offbtime'] = fl0['btime'] + fl0['uptime']

    # Get all rows from file1 where btime is greather than file0 offbtime
    conn1.execute('select rowid as startup, * from tuptime where btime > ' + str(fl0['offbtime']))
    db_rows = conn1.fetchall()

    print('\nRows to add from new file: \t' + str(len(db_rows)))

    # Insert rows on file0
    for row in db_rows:

        # At first row on file1, set offbtime, endst and downtime to last row on file0
        if row == db_rows[0]:

            print(' Fix shutdown values on row: ' + str(fl0['startup']))
            fl0['downtime'] = row['btime'] - fl0['offbtime']
            conn0.execute('update tuptime set offbtime = ' + str(fl0['offbtime']) + ', endst = 1, downtime = ' + str(fl0['downtime']) +
                          ' where rowid = (select max(rowid) from tuptime)')
            if arg.verbose:
                print('\toffbtime = ' + str(fl0['offbtime']))
                print('\tendst = 1')
                print('\tdowntime = ' + str(fl0['downtime']))

        # Add registers to file0
        print(' Adding startup row: ' + str(row['startup']))
        conn0.execute('insert into tuptime values (?,?,?,?,?,?,?,?)',
                      (str(row['btime']), str(row['uptime']), str(row['rntime']), str(row['slptime']), str(row['offbtime']), str(row['endst']), str(row['downtime']), str(row['kernel'])))
        if arg.verbose:
            print('\tbtime = ' + str(row['btime']))
            print('\tuptime = ' + str(row['uptime']))
            print('\trntime = ' + str(row['rntime']))
            print('\tslptime = ' + str(row['slptime']))
            print('\toffbtime = ' + str(row['offbtime']))
            print('\tendst = ' + str(row['endst']))
            print('\tdowntime = ' + str(row['downtime']))
            print('\tkernel = ' + str(row['kernel']))

        # At last row, empty offbtime, endst and downtime
        if row == db_rows[-1]:
            print(' Fix shutdown values on last row')
            conn0.execute('update tuptime set offbtime = NULL, endst = 0, downtime = NULL where rowid = (select max(rowid) from tuptime)')

    db_conn0.commit()

    # Print raw rows
    conn0.execute('select rowid as startup, * from tuptime')
    db_rows = conn0.fetchall()
    print('\nDestination rows after: \t' + str(len(db_rows)))
    if arg.verbose:
        for row in db_rows:
            print("\t", end='')
            print(dict(row))

    db_conn0.close()
    db_conn1.close()
    print('\nDone.')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
This script modify the "Startup Date", "Shutdown Date" and "End Status" on
Tuptime database keeping the other values around in sync.

Increase 60 secs the startup date on register number 1:
    tuptime_modify.py -c startup -r 1 -s 60

Decrease 100 secs the startup date on register number 4 on other file:
    tuptime_modify.py -c startup -r 4 -s -100 -f /tmp/test.db

Increase 30 secs the shutdown date on register number 12:
    tuptime_modify.py -c shutdown -r 12 -s 60

Decrease 300 secs the shutdown date on register number 47 with verbose:
    tuptime_modify.py -c shutdown -r 47 -s -300 -v

Swich end status value on register 54:
    tuptime_modify.py -c endst -r 54
'''

import sys, argparse, locale, signal, logging, sqlite3, tempfile
from shutil import copyfile

DB_FILE = '/var/lib/tuptime/tuptime.db'
DATE_FORMAT = '%X %x'
__version__ = '1.0.0'

# Terminate when SIGPIPE signal is received
signal.signal(signal.SIGPIPE, signal.SIG_DFL)

# Set locale to the users default settings (LANG env. var)
locale.setlocale(locale.LC_ALL, '')


def get_arguments():
    """Get arguments from command line"""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-c', '--change',
        dest='change',
        default=False,
        action='store',
        type=str,
        required=True,
        choices=['startup', 'shutdown', 'endst'],
        help='change startup or shutdown date time [<startup|shutdown|endst>]'
    )
    parser.add_argument(
        '-f', '--filedb',
        dest='db_file',
        default=DB_FILE,
        action='store',
        help='database file (' + DB_FILE + ')',
        metavar='FILE'
    )
    parser.add_argument(
        '-n', '--nobackup',
        dest='nobackup',
        default=False,
        action='store_true',
        help='avoid create backup db file'
    )
    parser.add_argument(
        '-r', '--register',
        dest='register',
        default=False,
        action='store',
        type=int,
        required=True,
        help='startup register to modify'
    )
    parser.add_argument(
        '-s', '--seconds',
        dest='seconds',
        default=0,
        action='store',
        type=int,
        help='seconds to add or remove (+/-)'
    )
    parser.add_argument(
        '-v', '--verbose',
        dest='verbose',
        default=False,
        action='store_true',
        help='verbose output'
    )
    arg = parser.parse_args()

    if arg.verbose:
        logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
        logging.info('Version: %s', (__version__))

    if arg.change == 'endst' and arg.seconds:
        parser.error('Operator \"seconds\" can\'t be combined with \"endst\"')

    logging.info('Arguments: %s', str(vars(arg)))
    return arg


def backup_dbf(arg):
    """Backup db file before modify"""

    try:
        tmp_file = tempfile.NamedTemporaryFile(prefix="tuptime_", suffix=".db").name

        copyfile(arg.db_file, tmp_file)
        print('Backup file:\t' + tmp_file)
    except Exception as exp:
        logging.error('Can\'t create backup file. %s', str(exp))
        sys.exit(-1)


def fix_endst(arg, reg, conn, modt, orgt):
    """Swich end status register"""

    # Swich between values 1 to 0 and 0 to 1
    modt['endst'] = 1 - orgt['endst']

    print('\t   modified\tendst: ' + str(modt['endst']))

    # Update values
    conn.execute('update tuptime set endst = ' + str(modt['endst']) + ' where rowid = ' + str(reg['target']))


def fix_shutdown(arg, reg, conn, modt, orgt):
    """Modify shutdown date register"""

    # Last row have None values
    if orgt['offbtime'] is None and orgt['downtime'] is None:
        modt['offbtime'] = None
        modt['downtime'] = None
    else:
        modt['offbtime'] = orgt['offbtime'] + arg.seconds
        modt['downtime'] = orgt['downtime'] - arg.seconds
    modt['uptime'] = orgt['uptime'] + arg.seconds
    
    if orgt['rntime'] + arg.seconds > 0:
        modt['rntime'] = orgt['rntime'] + arg.seconds
        modt['slptime'] = orgt['slptime']
    elif orgt['slptime'] + arg.seconds > 0:
        modt['rntime'] = orgt['rntime']
        modt['slptime'] = orgt['slptime'] + arg.seconds
    else:
        modt['rntime'] = modt['uptime']
        modt['slptime'] = 0

    print('\t   modified\tbtime:    n/a     | uptime: ' + str(modt['uptime']) + ' | rntime: ' + str(modt['rntime']) + ' | slptime: ' + str(modt['slptime']) + ' | offbtime: ' + str(modt['offbtime']) + ' | downtime: ' + str(modt['downtime']))

    # Limit check
    if modt['downtime'] is None or modt['downtime'] < 0 or \
       modt['offbtime'] is None or modt['offbtime'] < 0 or \
       modt['uptime'] < 0 or modt['rntime'] < 0 or modt['slptime'] < 0:
        logging.error('modified values can\'t be None or lower than 0')
        sys.exit(-1)

    # Update values
    conn.execute('update tuptime set uptime = ' + str(modt['uptime']) + ' where rowid = ' + str(reg['target']))
    conn.execute('update tuptime set rntime = ' + str(modt['rntime']) + ' where rowid = ' + str(reg['target']))
    conn.execute('update tuptime set slptime = ' + str(modt['slptime']) + ' where rowid = ' + str(reg['target']))
    conn.execute('update tuptime set downtime = ' + str(modt['downtime']) + ' where rowid = ' + str(reg['target']))
    conn.execute('update tuptime set offbtime = ' + str(modt['offbtime']) + ' where rowid = ' + str(reg['target']))


def fix_startup(arg, reg, conn, modt, orgt, modp, orgp):
    """Modify startup date register"""

    modt['btime'] = orgt['btime'] + arg.seconds
    modt['uptime'] = orgt['uptime'] - arg.seconds

    if orgt['rntime'] - arg.seconds > 0:
        modt['rntime'] = orgt['rntime'] - arg.seconds
        modt['slptime'] = orgt['slptime']
    elif orgt['slptime'] - arg.seconds > 0:
        modt['rntime'] = orgt['rntime']
        modt['slptime'] = orgt['slptime'] - arg.seconds
    else:
        modt['rntime'] = modt['uptime']
        modt['slptime'] = 0

    print('\t   modified\tbtime: ' + str(modt['btime']) + ' | uptime: ' + str(modt['uptime']) + ' | rntime: ' + str(modt['rntime']) + ' | slptime: ' + str(modt['slptime']) + ' | offbtime: ----n/a--- | downtime: --n/a-- ')

    # Limit check
    if modt['uptime'] < 0 or modt['rntime'] < 0 or modt['slptime'] < 0:
        logging.error('modified values can\'t be lower than 0')
        sys.exit(-1)

    # First row don't have previous register to modify
    if reg['prev'] > 0:

        # Get values from previous register
        conn.execute('select downtime from tuptime where rowid = ' + str(reg['prev']))
        orgp['downtime'] = conn.fetchone()[0]

        modp['downtime'] = orgp['downtime'] + arg.seconds

        print('\tTarget row-1 \'' + str(reg['prev']) + '\':')
        print('\t   original\tdowntime: ' + str(orgp['downtime']))
        print('\t   modified\tdowntime: ' + str(modp['downtime']))

        # Limit check
        if modp['downtime'] < 0:
            logging.error('downtime can\'t be lower than 0')
            sys.exit(-1)

        conn.execute('update tuptime set downtime = ' + str(modp['downtime']) + ' where rowid = ' + str(reg['prev']))

    # Update values
    conn.execute('update tuptime set btime = ' + str(modt['btime']) + ' where rowid = ' + str(reg['target']))
    conn.execute('update tuptime set uptime = ' + str(modt['uptime']) + ' where rowid = ' + str(reg['target']))
    conn.execute('update tuptime set rntime = ' + str(modt['rntime']) + ' where rowid = ' + str(reg['target']))
    conn.execute('update tuptime set slptime = ' + str(modt['slptime']) + ' where rowid = ' + str(reg['target']))


def main():

    arg = get_arguments()
    orgt = {}  # Original target value
    orgp = {}  # Original previous (target-1) value
    modt = {}  # Modified target value
    modp = {}  # Modified previous (target-1) value
    reg = {'target': arg.register, 'prev': (arg.register - 1)}  # Target register to modify and previous one

    print('Target file:\t' + arg.db_file)
    if not arg.nobackup:
        backup_dbf(arg)

    db_conn = sqlite3.connect(arg.db_file)
    db_conn.row_factory = sqlite3.Row
    conn = db_conn.cursor()

    # Check if DB have the old format
    columns = [i[1] for i in conn.execute('PRAGMA table_info(tuptime)')]
    if 'rntime' and 'slptime' not in columns:
        logging.error('DB format outdated')
        sys.exit(-1)

    # Print raw rows
    if arg.verbose:
        print('\nTarget registers before: ')
        conn.execute('select rowid as startup, * from tuptime where rowid between ' + str(reg['prev']) + ' and ' + str(reg['target']))
        db_rows = conn.fetchall()
        for row in db_rows:
            print("\t", end='')
            print(dict(row))

    # Validate register to modify
    conn.execute('select max(rowid) from tuptime')
    max_register = conn.fetchone()[0]
    if reg['target'] > max_register or reg['target'] < 1:
        logging.error('Invalid register to modify. Out of range')
        sys.exit(-1)

    # Get values from target row
    conn.execute('select btime, uptime, rntime, slptime, offbtime, endst, downtime from tuptime where rowid = ' + str(reg['target']))
    orgt['btime'], orgt['uptime'], orgt['rntime'], orgt['slptime'], orgt['offbtime'], orgt['endst'], orgt['downtime'] = conn.fetchone()

    print('\nValues:')
    print('\tTarget row \'' + str(reg['target']) + '\':')

    # Choose value to modify: endst / startup / shutdown
    if arg.change == 'endst':

        print('\t   original\tendst: ' + str(orgt['endst']))
        fix_endst(arg, reg, conn, modt, orgt)

    else:

        print('\t   original\tbtime: ' + str(orgt['btime']) + ' | uptime: ' + str(orgt['uptime']) + ' | rntime: ' + str(orgt['rntime']) + ' | slptime: ' + str(orgt['slptime']) + ' | offbtime: ' + str(orgt['offbtime']) + ' | downtime: ' + str(orgt['downtime']))

        if arg.change == 'startup':
            fix_startup(arg, reg, conn, modt, orgt, modp, orgp)
        if arg.change == 'shutdown':
            fix_shutdown(arg, reg, conn, modt, orgt)

    db_conn.commit()

    # Print raw rows
    if arg.verbose:
        print('\nTarget registers after: ')
        conn.execute('select rowid as startup, * from tuptime where rowid between ' + str(reg['prev']) + ' and ' + str(reg['target']))
        db_rows = conn.fetchall()
        for row in db_rows:
            print("\t", end='')
            print(dict(row))

    db_conn.close()
    print('\nDone.')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Sample plot report from the info that the tuptime command report"""

import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime, timedelta
import dateutil.parser
import subprocess, csv, argparse, tempfile


def get_arguments():
    """Get arguments from command line"""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-b', '--bdate',
        dest='bdate',
        action='store',
        help='begin date to plot, format:"Y-m-d"',
        type=str
    )
    parser.add_argument(
        '-e', '--edate',
        dest='edate',
        action='store',
        help='end date to plot, format:"Y-m-d". Default edate is today.',
        type=str
    )
    parser.add_argument(
        '-f', '--filedb',
        dest='dbfile',
        default=None,
        action='store',
        help='database file'
    )
    parser.add_argument(
        '-H', '--height',
        dest='height',
        default=6,
        action='store',
        help='window height.',
        type=int
    )
    parser.add_argument(
        '-p', '--pastdays',
        dest='pdays',
        default=7,
        action='store',
        help='past days before edate to plot, will be ignored if set bdate (default is 7).',
        type=int
    )
    parser.add_argument(
        '-W', '--width',
        dest='width',
        default=8,
        action='store',
        help='window width.',
        type=int
    )
    arg = parser.parse_args()
    return arg


def date_check(arg):
    """Check and clean dates"""

    # Set user or default begind date. Days ago...
    if arg.bdate:
        begin_date = dateutil.parser.parse(arg.bdate)
    else:
        begin_date = datetime.today() - timedelta(days=arg.pdays)
        print('Default begin:\t' + str(arg.pdays) + ' days ago')

    # Set user or default end date.
    if arg.edate:
        end_date = dateutil.parser.parse(arg.edate)
    else:
        end_date = datetime.today()
        print('Default end:\tnow')

    # Adjust date to the start or end time range and set the format
    begin_date = begin_date.replace(hour=0, minute=0, second=0).strftime("%Y-%m-%d %H:%M:%S")
    end_date = end_date.replace(hour=23, minute=59, second=59).strftime("%Y-%m-%d %H:%M:%S")

    print('Begin date:\t' + str(begin_date))
    print('End date:\t' + str(end_date))

    return([begin_date, end_date])


def date_range(date_limits):
    """Get the range of dates to apply"""

    dlimit = []  # Store date range in human date
    ranepo = []  # Store date range in epoch
    xlegend = []  # legend to x axis

    dstart = dateutil.parser.parse(date_limits[0])  # Start date
    dend = dateutil.parser.parse(date_limits[1])  # End date

    # Split time range in days and finally add the rest (minnor than one day)
    while dstart <= dend:
        dlimit.append(dstart)
        dstart += timedelta(days=1)
    dlimit.append(dend)

    print('Date splits:\t' + str(len(dlimit)))

    # Create list with lists of two epoch dates, begin and end for each split
    reg = 1
    while reg < len(dlimit):
        ranepo.append([int(dlimit[reg-1].timestamp()), int(dlimit[reg].timestamp())])
        xlegend.append(datetime.fromtimestamp(dlimit[reg-1].timestamp()).strftime('%Y-%m-%d'))
        reg += 1

    print('Ranges on list:\t' + str(len(ranepo)))

    return(ranepo, xlegend)


def main():
    """Core logic"""

    arg = get_arguments()
    date_limits = date_check(arg)
    date_list, xlegend = date_range(date_limits)

    # Query to Tuptime with each range
    nran = 0  # Range number
    daysplt = []  # List for all day splits with their events.
    max_events = 0  # Max events in a daysplit
    ftemp = tempfile.NamedTemporaryFile().name  # File to store Tuptime csv

    # Iterate over each element in (since, until) list
    while nran < len(date_list):
        ar1 = str(int(date_list[nran][0]))  # Timestamp arg tsince
        ar2 = str(int(date_list[nran][1]))  # timestamp arg tuntil
        nran += 1  # Number of range

        # Save Tuptime csv output to file
        with open(ftemp, "wb", 0) as out:
            if arg.dbfile:  # If a file is passed, avoid update it
                subprocess.call(["tuptime", "-lsc", "--tsince", ar1, "--tuntil", ar2, "-f", arg.dbfile, "-n"], stdout=out)
            else:
                subprocess.call(["tuptime", "-lsc", "--tsince", ar1, "--tuntil", ar2], stdout=out)

        # Get csv file
        l_events = []  # List for events in each daysplt
        with open(ftemp) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',')
            bad = None  # Register bad shutdown

            for row in csv_reader:
                l_row = []  # List for events in csv rows

                # Control how was the shutdown
                if (row[0] == 'Shutdown') and (row[1] == 'BAD'):
                    bad = True

                # This create a list with (uptime, downtime ok, downtime bad)
                if (row[0] == 'Uptime') or (row[0] == 'Downtime'):

                    if row[0] == 'Uptime':
                        # Add 0 or value to first position
                        l_row.append(int(round(float(row[1]), 0)))
                    else:
                        l_row.append(0)

                    if row[0] == 'Downtime':
                        # Add (0, value) or (value, 0) to the end
                        if bad is True:
                            l_row.extend((0, int(round(float(row[1]), 0))))
                        else:
                            l_row.extend((int(round(float(row[1]), 0)), 0))
                        bad = False  # Reset false
                    else:
                        l_row.extend((0, 0))

                    # Add to events list
                    l_events.append(l_row)

            # Add to daysplt list
            daysplt.append(l_events)

            print('Got range --->\t' + str(nran) + ' with ' + str(len(l_events)) + ' events')

            # Catch large l_events list
            if len(l_events) > max_events:
                max_events = len(l_events)

    print('Ranges got:\t' + str(len(daysplt)))

    # Convert seconds to days to plot them
    for i in range(0, len(daysplt)):
        for k in range(0, len(daysplt[i])):
            daysplt[i][k][:] = [x / 3600 for x in daysplt[i][k]]

    # All days with equal number of events, fill with 0,0,0
    for i in range(0, len(daysplt)):
        while max_events > len(daysplt[i]):
            daysplt[i].append([0, 0, 0])

    # Stack events from each day (daysplt) based on their position
    #  upt, downt ok, downt bad, upt, downt ok, and so on...
    stack = []
    for k in range(0, max_events):
        for n in range(0, 3):
            position1 = []
            for u in range(0, len(daysplt)):
                position1.append(daysplt[u][k][n])
            stack.append(np.array(position1))

    print('Rows to stack:\t' + str(len(stack)))

    ind = np.arange(len(daysplt))    # the x locations for the groups
    width = 0.9
    color_list = ['green', 'grey', 'black']

    plt.figure(figsize=(arg.width, arg.height))
    # Plot events from their position and rotate color
    for i in range(0, len(stack)):
        plt.bar(ind, stack[i], width, color=color_list[i % len(color_list)], linewidth=0, bottom=np.sum(stack[:i], axis=0))

    plt.ylabel('Hours')
    plt.title('Tuptime bar chart')
    plt.xticks(ind, xlegend, rotation=15)
    plt.margins(y=0, x=0.01)
    plt.yticks(np.arange(0, 25, 2))
    plt.grid(color='w', linestyle='--', linewidth=0.5, axis='y')
    plt.tight_layout()
    cfig = plt.get_current_fig_manager()
    cfig.canvas.set_window_title("Tuptime")
    plt.show()

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Migrate Uptimed records to Tuptime"""

import os, sys, logging, sqlite3, argparse
from shutil import which, copyfile

# Script to migrate uptimed records to a sqlite file suitable to be
# used by tuptime.
# Adding records to an existing tuptime db is not supported.

# Copyright: gustavo panizzo <gfa@zumbi.com.ar> 2018-11-18
# License: GPL-2.0+
#
# V1.0: Gustavo Panizzo 2018-11-18
# V1.1: Ricardo Fraile 2018-11-21
# V1.2: Ricardo Fraile 2019-03-24
# V1.3: Ricardo Fraile 2019-10-19
__version__ = '1.3'

# Default Tuptime db location
TUPT_DBF = '/var/lib/tuptime/tuptime.db'

# Temporary file for Tuptime db
TUPT_TMP_DBF = '/tmp/tuptime-%s.db' % os.getpid()

# Backup for the current Tuptime db, if exists
TUPT_BACKUP = '/var/tmp/tuptime-%s.db.backup' % os.getpid()

# Overwrite toggle
REPLACE_DBF = False

# Location of uptimed records file (format uptime:btime:kernel)
UPTIME_RECORDS_FILE = '/var/spool/uptimed/records'



def insert_row(row):
    """ inserts a uptimed record as a row """

    conn = sqlite3.connect(TUPT_TMP_DBF)
    _uptime = int(round(float(row[0]), 0))
    _rntime = int(round(float(row[0]), 0))
    _slptime = int(0)
    _btime = int(round(float(row[1]), 0))
    _kernel = row[2]
    if row[3]:
        _offbtime = int(round(float(row[3]), 0))
    else:
        _offbtime = None
    if row[4]:
        _downtime = int(round(float(row[4]), 0))
    else:
        _downtime = None
    _endst = int(row[5])
    params = [_btime, _uptime, _rntime, _slptime, _offbtime, _endst, _downtime, _kernel]
    print(str(params))

    conn.execute("INSERT INTO tuptime VALUES (?,?,?,?,?,?,?,?)", params)
    conn.commit()
    conn.close()



def create_db(db_file):
    """ creates a sqlite file and the tables required for tuptime. """

    print('+ Creating Tuptime temp db file: ' + str(db_file) + '\n')

    if os.path.isfile(db_file):
        logging.error('DB exists: ' + (db_file))
        raise SystemExit

    db_conn = sqlite3.connect(db_file)
    conn = db_conn.cursor()
    conn.execute('create table if not exists tuptime'
                 '(btime integer, uptime integer, rntime integer, slptime integer,'
                 'offbtime integer, endst integer, downtime integer, kernel text)')
    db_conn.commit()
    db_conn.close()



def read_uptime_records(file):
    """ returns uptimed records in a list """

    print('+ Reading Uptimed file: ' + UPTIME_RECORDS_FILE + '\n')

    # Read file and split records
    uptimed_records = []
    try:
        with open(file) as records:
            for line in records:
                line = line.rstrip()
                line = line.split(':')
                uptimed_records.append(line)
    except Exception as exc:
        sys.exit(exc)

    # Order by btime and reverse by uptime
    uptimed_records = sorted(uptimed_records, key=lambda x: (int(x[1]), -int(x[0])))

    # Exclude invalid records
    uptimed_clean_records = []
    offbtime = None
    for i, line in enumerate(uptimed_records):
        if i > 0:

            # If btime is lower than previous btime + uptime
            if int(line[1]) < int(offbtime):
                print('Exclude btime overlap --> ' + str(line))
                continue

            # If is negative
            if int(line[0]) < 0:  # uptime < 0
                print('Exclude uptime negative --> ' + str(line))
                continue

        offbtime = int(line[0]) + int(line[1])  # uptime + btime
        print(line)
        uptimed_clean_records.append(line)

    print('Total valid records: ' + str(len(uptimed_clean_records)) + '\n')

    if not len(uptimed_clean_records):
        logging.error('Any records to migrate from uptimed')
        raise SystemExit

    return uptimed_clean_records



def uptimed_to_tuptime(uptimed):
    """ converts records from uptimed format to tuptime's format """

    # Create tuptime content:
    #     uptime, btime, kernel + offbtime, downtime, endst
    newrecords = []
    for i, line in enumerate(uptimed):

        # Current offbtime
        offbtime = int(line[0]) + int(line[1])
        line.append(offbtime)

        if i > 0:
            # Previous downtime
            prev_downtime = int(line[1]) - int(newrecords[i-1][3])  # btime - prev.offbtime
            newrecords[i-1].append(prev_downtime)
            # Previous shutdown status
            newrecords[i-1].append('1')
        newrecords.append(line)

    # Reset last line
    newrecords[-1][3] = False  # offbtime
    newrecords[-1].append(False)  # downtime
    newrecords[-1].append('1')  # endst

    if not len(newrecords):
        logging.error('Any records to migrate to Tuptime')
        raise SystemExit

    return newrecords



def substitution():
    """ test db result and copy to default location """

    print('\n+ Checking Tuptime execution over temp db file \n')

    if os.system("tuptime -tf " + TUPT_TMP_DBF) != 0:
        logging.error('Tuptime temp execution fails.')
        raise SystemExit

    if REPLACE_DBF:
        print('\n+ Moving temp db file to default Tuptime location: ' + TUPT_DBF)
        try:
            copyfile(TUPT_TMP_DBF, TUPT_DBF)
        except Exception as excp:
            sys.exit(excp)

        print('\n+ Setting user and group\n')
        user = os.stat(os.path.dirname(TUPT_DBF)).st_uid
        group = os.stat(os.path.dirname(TUPT_DBF)).st_gid
        os.chown(TUPT_DBF, user, group)
    else:
        print("\n(run with \'--replace\' to apply changes)")

    print('+ Done \n')



def checks():
    """ validate requirements """

    # Arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--replace',
        dest='replace',
        action='store_true',
        default=False,
        help='Replace default Tuptime db file'
    )
    parser.add_argument(
        '-v', '--version',
        action='version',
        version='version ' + (__version__),
        help='show version'
    )
    arg = parser.parse_args()
    if arg.replace:
        global REPLACE_DBF
        REPLACE_DBF = True

    # Privileged user
    if os.geteuid() != 0:
        logging.error('Execute with a privileged user')
        raise SystemExit

    # Tuptime present
    if not which('tuptime'):
        logging.error('Tuptime not installed')
        raise SystemExit

    # Uptime file exists
    if not os.access(UPTIME_RECORDS_FILE, os.R_OK):
        logging.error('Can\'t read ' + UPTIME_RECORDS_FILE)
        raise SystemExit

    # Backup default Tuptime db file if exists
    if os.path.isfile(TUPT_DBF):
        try:
            copyfile(TUPT_DBF, TUPT_BACKUP)
        except Exception as excp:
            sys.exit(excp)
        print('Backup current Tuptime db file to ' + TUPT_BACKUP + '\n')



def main():
    """ Main migration steps """

    print('\n####  Uptimed to Tuptime migration script  ####\n')

    checks()
    uptimed_records = read_uptime_records(UPTIME_RECORDS_FILE)
    newrecords = uptimed_to_tuptime(uptimed_records)
    create_db(TUPT_TMP_DBF)
    for record in newrecords:
        insert_row(record)
    substitution()


if __name__ == "__main__":
    main()
